%% ===================================================================
%% @author Krzysztof Trzepla
%% @copyright (C): 2014 ACK CYFRONET AGH
%% This software is released under the MIT license
%% cited in 'LICENSE.txt'.
%% @end
%% ===================================================================
%% @doc This module contains n2o website code.
%% This page allows to view space details.
%% @end
%% ===================================================================
-module(page_space_details).

-include("gui_modules/common.hrl").
-include_lib("ctool/include/logging.hrl").
-include_lib("ctool/include/oz/oz_users.hrl").
-include_lib("ctool/include/oz/oz_spaces.hrl").
-include_lib("ctool/include/oz/oz_providers.hrl").

-export([main/0, event/1, comet_loop/1]).

%% Common page CCS styles
-define(CONTENT_COLUMN_STYLE, <<"padding-right: 0">>).
-define(NAVIGATION_COLUMN_STYLE, <<"border-left-width: 0; width: 20px; padding-left: 0;">>).
-define(PARAGRAPH_STYLE, <<"margin: 0 auto;">>).

%% Comet process pid
-define(COMET_PID, comet_pid).

%% Comet process state
-define(STATE, comet_state).
-record(?STATE, {providers_details, users_details}).

%% ====================================================================
%% API functions
%% ====================================================================

%% main/0
%% ====================================================================
%% @doc Template points to the template file, which will be filled with content.
%% @end
-spec main() -> Result when
    Result :: #dtl{}.
%% ====================================================================
main() ->
    case gui_ctx:user_logged_in() of
        true ->
            case provider_logic:get_provider_id() of
                undefined ->
                    page_error:redirect_with_error(?UNREGISTERED_PROVIDER_ERROR),
                    #dtl{file = "bare", app = ?APP_NAME, bindings = [{title, <<"">>}, {body, <<"">>}, {custom, <<"">>}]};
                _ ->
                    case gui_ctx:url_param(<<"id">>) of
                        undefined ->
                            page_error:redirect_with_error(?SPACE_NOT_FOUND_ERROR),
                            #dtl{file = "bare", app = ?APP_NAME, bindings = [{title, <<"">>}, {body, <<"">>}, {custom, <<"">>}]};
                        Id ->
                            SpaceId = str_utils:to_binary(Id),
                            case oz_providers:get_space_details(provider, SpaceId) of
                                {ok, SpaceDetails} ->
                                    #dtl{file = "bare", app = ?APP_NAME, bindings = [{title, title()}, {body, body(SpaceDetails)}, {custom, <<"">>}]};
                                Other ->
                                    ?error("Cannot get details of Space with ID ~p: ~p", [SpaceId, Other]),
                                    page_error:redirect_with_error(?SPACE_PERMISSION_DENIED_ERROR),
                                    #dtl{file = "bare", app = ?APP_NAME, bindings = [{title, <<"">>}, {body, <<"">>}, {custom, <<"">>}]}
                            end
                    end
            end;
        false ->
            gui_jq:redirect_to_login(),
            #dtl{file = "bare", app = ?APP_NAME, bindings = [{title, <<"">>}, {body, <<"">>}, {custom, <<"">>}]}
    end.


%% title/0
%% ====================================================================
%% @doc This will be placed instead of {{title}} tag in template.
%% @end
-spec title() -> Result when
    Result :: binary().
%% ====================================================================
title() ->
    <<"Space details">>.


%% body/0
%% ====================================================================
%% @doc This will be placed instead of {{body}} tag in template.
%% @end
-spec body(SpaceDetails :: #space_details{}) -> Result when
    Result :: #panel{}.
%% ====================================================================
body(SpaceDetails) ->
    Header = onepanel_gui_utils_adapter:top_menu(spaces_tab, spaces_dashboard_link, [], true),
    Main = #panel{
        style = <<"margin-top: 2em; text-align: center;">>,
        body = [
            #h6{
                style = <<"font-size: x-large; margin-bottom: 2em;">>,
                body = <<"Space details">>
            },
            space_details_table(SpaceDetails) |
            lists:map(fun(TableId) ->
                #table{
                    class = <<"table table-bordered table-striped">>,
                    style = <<"width: 50%; margin: 0 auto; margin-top: 3em; table-layout: fixed;">>,
                    body = #tbody{
                        id = TableId,
                        style = <<"display: none;">>
                    }
                }
            end, [<<"providers_table">>, <<"users_table">>])
        ]
    },
    onepanel_gui_utils:body(Header, Main).


%% space_details_table/1
%% ====================================================================
%% @doc Renders the body of space details table
%% @end
-spec space_details_table(SpaceDetails :: #space_details{}) -> Result when
    Result :: #table{}.
%% ====================================================================
space_details_table(#space_details{id = SpaceId, name = SpaceName}) ->
    DescriptionStyle = <<"border-width: 0; vertical-align: top; text-align: right; padding: 1em 1em;">>,
    MainStyle = <<"border-width: 0;  text-align: left; padding: 1em 1em;">>,
    #table{
        style = <<"border-width: 0; width: 50%; margin: 0 auto;">>,
        body = lists:map(fun({DetailName, DetailValue}) ->
            #tr{
                cells = [
                    #td{
                        style = DescriptionStyle,
                        body = #label{
                            class = <<"label label-large label-inverse">>,
                            style = <<"cursor: auto;">>,
                            body = DetailName
                        }
                    },
                    #td{
                        style = MainStyle,
                        body = #p{
                            style = ?PARAGRAPH_STYLE,
                            body = DetailValue
                        }
                    }
                ]
            }
        end, [
            {<<"Name">>, SpaceName},
            {<<"Space ID">>, SpaceId}
        ])
    }.


%% providers_table_collapsed/1
%% ====================================================================
%% @doc Renders collapsed providers details table.
%% @end
-spec providers_table_collapsed(ProvidersDetails :: [{Id :: binary(), ProviderDetails :: #provider_details{}}]) -> Result when
    Result :: [#tr{}].
%% ====================================================================
providers_table_collapsed(ProvidersDetails) ->
    TableName = <<"Providers">>,
    NavigationBody = onepanel_gui_utils:expand_button(<<"Expand All">>, {message, expand_providers_table}),
    RenderRowFunction = fun provider_row_collapsed/2,
    table(ProvidersDetails, TableName, NavigationBody, RenderRowFunction).


%% providers_table_expanded/1
%% ====================================================================
%% @doc Renders expanded providers details table.
%% @end
-spec providers_table_expanded(ProvidersDetails :: [{Id :: binary(), ProviderDetails :: #provider_details{}}]) -> Result when
    Result :: [#tr{}].
%% ====================================================================
providers_table_expanded(ProvidersDetails) ->
    TableName = <<"Providers">>,
    NavigationBody = onepanel_gui_utils:collapse_button(<<"Collapse All">>, {message, collapse_providers_table}),
    RenderRowFunction = fun provider_row_expanded/2,
    table(ProvidersDetails, TableName, NavigationBody, RenderRowFunction).


%% users_table_collapsed/1
%% ====================================================================
%% @doc Renders collapsed users details table.
%% @end
-spec users_table_collapsed(UsersDetails :: [{Id :: binary(), UserDetails :: #user_details{}}]) -> Result when
    Result :: [#tr{}].
%% ====================================================================
users_table_collapsed(UsersDetails) ->
    TableName = <<"Users">>,
    NavigationBody = onepanel_gui_utils:expand_button(<<"Expand All">>, {message, expand_users_table}),
    RenderRowFunction = fun user_row_collapsed/2,
    table(UsersDetails, TableName, NavigationBody, RenderRowFunction).


%% users_table_expanded/1
%% ====================================================================
%% @doc Renders expanded users details table.
%% @end
-spec users_table_expanded(UsersDetails :: [{Id :: binary(), UserDetails :: #user_details{}}]) -> Result when
    Result :: [#tr{}].
%% ====================================================================
users_table_expanded(UsersDetails) ->
    TableName = <<"Users">>,
    NavigationBody = onepanel_gui_utils:collapse_button(<<"Collapse All">>, {message, collapse_users_table}),
    RenderRowFunction = fun user_row_expanded/2,
    table(UsersDetails, TableName, NavigationBody, RenderRowFunction).


%% table/4
%% ====================================================================
%% @doc Renders details table.
%% @end
-spec table(Details :: [{Id :: binary(), Details :: #provider_details{} | #user_details{}}], TableName :: binary(), NavigationBody :: #link{}, RenderRowFunction :: function()) -> Result when
    Result :: [#tr{}].
%% ====================================================================
table(Details, TableName, NavigationBody, RenderRowFunction) ->
    Header = #tr{
        cells = [
            #th{
                style = <<"font-size: large;">>,
                body = TableName
            },
            #th{
                style = ?NAVIGATION_COLUMN_STYLE,
                body = NavigationBody
            }
        ]
    },

    Rows = lists:foldl(fun({RowId, RowDetails}, RowsAcc) ->
        [#tr{
            id = RowId,
            cells = RenderRowFunction(RowId, RowDetails)
        } | RowsAcc]
    end, [], Details),

    [Header | Rows].


%% provider_row_collapsed/2
%% ====================================================================
%% @doc Renders collapsed provider details row.
%% @end
-spec provider_row_collapsed(RowId :: binary(), ProviderDetails :: #provider_details{}) -> Result when
    Result :: [#td{}].
%% ====================================================================
provider_row_collapsed(RowId, #provider_details{id = ProviderId, name = ProviderName} = ProviderDetails) ->
    NavigationBody = onepanel_gui_utils:expand_button({message, {expand_provider_row, RowId, ProviderDetails}}),
    row_collapsed(ProviderId, ProviderName, NavigationBody).


%% provider_row_expanded/2
%% ====================================================================
%% @doc Renders expanded provider details row.
%% @end
-spec provider_row_expanded(RowId :: binary(), ProviderDetails :: #provider_details{}) -> Result when
    Result :: [#td{}].
%% ====================================================================
provider_row_expanded(RowId, #provider_details{id = ProviderId, name = ProviderName, redirection_point = RedirectionPoint, urls = URLs} = ProviderDetails) ->
    Details = [
        {<<"Name">>, #p{style = ?PARAGRAPH_STYLE, body = ProviderName}},
        {<<"Provider ID">>, #p{style = ?PARAGRAPH_STYLE, body = ProviderId}},
        {<<"URLs">>, #list{
            style = <<"list-style-type: none; margin: 0 auto;">>,
            body = lists:map(fun(URL) ->
                #li{body = #p{
                    style = ?PARAGRAPH_STYLE,
                    body = URL}
                }
            end, lists:sort(URLs))
        }},
        {<<"Redirection point">>, #p{style = ?PARAGRAPH_STYLE, body = RedirectionPoint}}
    ],
    NavigationBody = onepanel_gui_utils:collapse_button({message, {collapse_provider_row, RowId, ProviderDetails}}),
    row_expanded(Details, NavigationBody).


%% user_row_collapsed/2
%% ====================================================================
%% @doc Renders collapsed user details row.
%% @end
-spec user_row_collapsed(RowId :: binary(), UserDetails :: #user_details{}) -> Result when
    Result :: [#td{}].
%% ====================================================================
user_row_collapsed(RowId, #user_details{id = UserId, name = UserName} = UserDetails) ->
    NavigationBody = onepanel_gui_utils:expand_button({message, {expand_user_row, RowId, UserDetails}}),
    row_collapsed(UserId, UserName, NavigationBody).


%% user_row_expanded/2
%% ====================================================================
%% @doc Renders expanded user details row.
%% @end
-spec user_row_expanded(RowId :: binary(), UserDetails :: #user_details{}) -> Result when
    Result :: [#td{}].
%% ====================================================================
user_row_expanded(RowId, #user_details{id = UserId, name = UserName} = UserDetails) ->
    Details = [
        {<<"Name">>, #p{style = ?PARAGRAPH_STYLE, body = UserName}},
        {<<"User ID">>, #p{style = ?PARAGRAPH_STYLE, body = UserId}}
    ],
    NavigationBody = onepanel_gui_utils:collapse_button({message, {collapse_user_row, RowId, UserDetails}}),
    row_expanded(Details, NavigationBody).


%% row_collapsed/2
%% ====================================================================
%% @doc Renders collapsed details row.
%% @end
-spec row_collapsed(Id :: binary(), Name :: binary(), NavigationBody :: term()) -> Result when
    Result :: [#td{}].
%% ====================================================================
row_collapsed(Id, Name, NavigationBody) ->
    [
        #td{
            style = ?CONTENT_COLUMN_STYLE,
            body = #p{
                style = ?PARAGRAPH_STYLE,
                body = <<"<b>", Name/binary, "</b> (", Id/binary, ")">>
            }
        },
        #td{
            style = ?NAVIGATION_COLUMN_STYLE,
            body = NavigationBody
        }
    ].


%% row_expanded/2
%% ====================================================================
%% @doc Renders expanded details row.
%% @end
-spec row_expanded([{DetailName :: binary(), DetailBody :: term()}], NavigationBody :: term()) -> Result when
    Result :: [#td{}].
%% ====================================================================
row_expanded(Details, NavigationBody) ->
    [
        #td{
            style = ?CONTENT_COLUMN_STYLE,
            body = #table{
                style = <<"border-width: 0; width: 100%; border-collapse: inherit;">>,
                body = lists:map(fun({DetailName, DetailBody}) ->
                    #tr{
                        cells = [
                            #td{
                                style = <<"border-width: 0; text-align: right; width: 10%; padding-left: 0; padding-right: 0;">>,
                                body = #label{
                                    style = <<"margin: 0 auto; cursor: auto;">>,
                                    class = <<"label label-large label-inverse">>,
                                    body = DetailName
                                }
                            },
                            #td{
                                style = <<"border-width: 0;  text-align: left; padding-left: 1em; width: 90%;">>,
                                body = DetailBody
                            }
                        ]
                    }
                end, Details)
            }
        },
        #td{
            style = ?NAVIGATION_COLUMN_STYLE,
            body = NavigationBody
        }
    ].


%% ====================================================================
%% Events handling
%% ====================================================================

%% comet_loop/1
%% ====================================================================
%% @doc Handles spaces management actions.
%% @end
-spec comet_loop(State :: #?STATE{}) -> Result when
    Result :: {error, Reason :: term()}.
%% ====================================================================
comet_loop({error, Reason}) ->
    {error, Reason};

comet_loop(#?STATE{providers_details = ProvidersDetails, users_details = UsersDetails} = State) ->
    NewState = try
                   receive
                       render_tables ->
                           gui_jq:update(<<"providers_table">>, providers_table_collapsed(ProvidersDetails)),
                           gui_jq:fade_in(<<"providers_table">>, 500),
                           gui_jq:update(<<"users_table">>, users_table_collapsed(UsersDetails)),
                           gui_jq:fade_in(<<"users_table">>, 500),
                           State;

                       Event ->
                           case Event of
                               collapse_providers_table ->
                                   gui_jq:update(<<"providers_table">>, providers_table_collapsed(ProvidersDetails));
                               expand_providers_table ->
                                   gui_jq:update(<<"providers_table">>, providers_table_expanded(ProvidersDetails));
                               {collapse_provider_row, RowId, ProviderDetails} ->
                                   gui_jq:update(RowId, provider_row_collapsed(RowId, ProviderDetails));
                               {expand_provider_row, RowId, ProviderDetails} ->
                                   gui_jq:update(RowId, provider_row_expanded(RowId, ProviderDetails));
                               collapse_users_table ->
                                   gui_jq:update(<<"users_table">>, users_table_collapsed(UsersDetails));
                               expand_users_table ->
                                   gui_jq:update(<<"users_table">>, users_table_expanded(UsersDetails));
                               {collapse_user_row, RowId, UserDetails} ->
                                   gui_jq:update(RowId, user_row_collapsed(RowId, UserDetails));
                               {expand_user_row, RowId, UserDetails} ->
                                   gui_jq:update(RowId, user_row_expanded(RowId, UserDetails));
                               _ ->
                                   ok
                           end,
                           State

                   after ?COMET_PROCESS_RELOAD_DELAY ->
                       State
                   end
               catch Type:Message ->
                   ?error_stacktrace("Comet process exception: ~p:~p", [Type, Message]),
                   onepanel_gui_utils:message(error, <<"There has been an error in comet process. Please refresh the page.">>),
                   {error, Message}
               end,
    gui_jq:wire(<<"$('#main_spinner').delay(300).hide(0);">>, false),
    gui_comet:flush(),
    ?MODULE:comet_loop(NewState).


%% event/1
%% ====================================================================
%% @doc Handles page events.
%% @end
-spec event(Event :: term()) -> no_return().
%% ====================================================================
event(init) ->
    try
        SpaceId = str_utils:to_binary(gui_ctx:url_param(<<"id">>)),

        GetDetailsFun = fun(Ids, Function, RowPrefix) ->
            lists:foldl(fun(Id, {Rows, It}) ->
                {ok, Details} = oz_spaces:Function(provider, SpaceId, Id),
                {
                    [{<<RowPrefix/binary, (integer_to_binary(It + 1))/binary>>, Details} | Rows],
                    It + 1
                }
            end, {[], 0}, Ids)
        end,

        {ok, ProviderIds} = oz_spaces:get_providers(provider, SpaceId),
        {ProvidersDetails, _} = GetDetailsFun(ProviderIds, get_provider_details, <<"provider_">>),
        {ok, UserIds} = oz_spaces:get_users(provider, SpaceId),
        {UsersDetails, _} = GetDetailsFun(UserIds, get_user_details, <<"user_">>),

        {ok, Pid} = gui_comet:spawn(fun() ->
            comet_loop(#?STATE{providers_details = ProvidersDetails, users_details = UsersDetails})
        end),
        put(?COMET_PID, Pid),
        Pid ! render_tables
    catch
        _:Reason ->
            ?error_stacktrace("Cannot initialize page ~p: ~p", [?MODULE, Reason]),
            gui_jq:hide(<<"main_spinner">>),
            onepanel_gui_utils:message(error, <<"Cannot fetch Space details.<br>Please try again later.">>)
    end;

event({message, Message}) ->
    get(?COMET_PID) ! Message,
    gui_jq:show(<<"main_spinner">>);

event({close_message, MessageId}) ->
    gui_jq:hide(MessageId);

event(terminate) ->
    ok.